"""Calculations endpoints namespace"""

from collections.abc import Sequence

from ..helpers import absolidix_json_decoder, raise_on_absolidix_error
from .base import BaseNamespace


class AbsolidixCalculationsNamespace(BaseNamespace):
    """Calculations endpoints namespace"""

    @raise_on_absolidix_error
    async def supported(self) -> Sequence[str]:
        "Get supported calculation engines"
        async with await self._client.request(
            method="GET",
            url=self._base_url / "supported",
            auth_required=False,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)



"""Authentication endpoints namespace"""

from ..dtos import AbsolidixAuthCredentialsRequestDTO, AbsolidixUserDTO
from ..helpers import absolidix_json_decoder, raise_on_absolidix_error
from .base import BaseNamespace


class AbsolidixV0AuthNamespace(BaseNamespace):
    """Authentication endpoints namespace"""

    @raise_on_absolidix_error
    async def login(self, email: str, password: str) -> bool:
        "Login"
        payload = AbsolidixAuthCredentialsRequestDTO(email=email, password=password)
        async with self._client.request(
            method="POST",
            url=self._base_url,
            json=payload,
        ) as resp:
            return resp.ok

    @raise_on_absolidix_error
    async def whoami(self) -> AbsolidixUserDTO:
        "Get self info"
        async with self._client.request(
            url=self._base_url, auth_required=self._auth_required
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)



"""Compatibility layer"""

import sys

if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import TypeAlias, TypeGuard
else:  # pragma: no cover
    from typing import TypeAlias, TypeGuard

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Concatenate, NotRequired, ParamSpec, TypedDict, Unpack
else:  # pragma: no cover
    from typing import Concatenate, NotRequired, ParamSpec, TypedDict, Unpack


__all__ = [
    "Concatenate",
    "NotRequired",
    "ParamSpec",
    "TypeAlias",
    "TypeGuard",
    "TypedDict",
    "Unpack",
]



"""Models"""

from .auth import (
    AbsolidixLocalUserAuth,
    AbsolidixNoAuth,
    AbsolidixTokenAuth,
    BaseAuthenticator,
)
from .base import AbsolidixBase
from .event import AbsolidixMessageEvent
from .hub import AbsolidixHub
from .subscription import AbsolidixSubscription, act_and_get_result_from_stream

__all__ = [
    "AbsolidixBase",
    "AbsolidixHub",
    "AbsolidixMessageEvent",
    "act_and_get_result_from_stream",
    "BaseAuthenticator",
    "AbsolidixLocalUserAuth",
    "AbsolidixNoAuth",
    "AbsolidixTokenAuth",
    "AbsolidixSubscription",
]



"Test AbsolidixAPI"

from typing import Union

import pytest
from aiohttp import web
from aiohttp.test_utils import TestServer
from yarl import URL

from absolidix_client import AbsolidixAPI, AbsolidixNoAuth
from absolidix_client.exc import (
    AbsolidixAsyncRuntimeWarning,
    AbsolidixConnectionException,
    AbsolidixException,
)
from tests.helpers import TestClient


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    return app


@pytest.fixture
async def aiohttp_client():
    "Create test client"
    app = await create_app()
    async with TestClient(TestServer(app)) as client:
        yield client


@pytest.fixture
async def base_url(aiohttp_client: TestClient) -> URL:
    "Return base url"
    return aiohttp_client.make_url("")


@pytest.mark.parametrize(
    "default_timeout, timeout, result",
    [
        (False, False, None),
        (False, None, None),
        (False, 1, 1),
        (None, False, None),
        (None, None, None),
        (None, 1, 1),
        (1, False, None),
        (1, None, 1),
        (1, 1, 1),
    ],
)
async def test_sync_ns_timeout(
    base_url: URL,
    default_timeout: Union[bool, None, int],
    timeout: Union[bool, None, int],
    result: Union[bool, None, int],
):
    "Test timeout guessing method"
    client = AbsolidixAPI(base_url, auth=AbsolidixNoAuth(), timeout=default_timeout)

    assert client.v0._get_timeout(timeout) == result  # pyright: ignore[reportPrivateUsage]


async def test_sync_in_async_runtime(base_url: URL):
    """
    Use sync API in an async runtime.
    It's impossible to connect to the test fake server
    because of runtime block by synchronous function call.
    So, we check that:
     - RuntimeError is absent (asyncio.run() cannot be called from a running event loop)
     - AbsolidixAsyncRuntimeWarning is present about misuse of API
     - timeout exception is present
    """
    client = AbsolidixAPI(base_url, auth=AbsolidixNoAuth(), timeout=0.0001)
    with pytest.warns(AbsolidixAsyncRuntimeWarning):
        with pytest.raises((AbsolidixConnectionException, AbsolidixException)):  # noqa: B908
            client.v0.auth.whoami()



"""User DTOs"""

from ..compat import NotRequired
from .base import AbsolidixTimestampsDTO


class AbsolidixUserDTO(AbsolidixTimestampsDTO):
    "User DTO"

    id: int
    first_name: NotRequired[str]
    last_name: NotRequired[str]
    email: NotRequired[str]
    email_verified: NotRequired[bool]
    role_label: NotRequired[str]
    role_slug: NotRequired[str]
    permissions: NotRequired[dict[str, str]]
    provider: NotRequired[str]



"Test method wrappers"

from collections.abc import Awaitable
from contextlib import nullcontext
from typing import Any, Callable, TypeVar

import pytest

from absolidix_client.exc import AbsolidixError
from absolidix_client.helpers import (
    raise_on_absolidix_error,
    raise_on_absolidix_error_in_event,
)
from tests.test_helpers_type_guards import (
    IS_ABSOLIDIX_ERROR_DTO_PAMAMS,
    IS_ABSOLIDIX_ERRORS_EVT_DTO_PARAMS,
)

T = TypeVar("T")


def mk_async_fun(x: T) -> Callable[[], Awaitable[T]]:
    "Async echo"

    async def run():
        return x

    return run


@pytest.mark.parametrize("x,expected", IS_ABSOLIDIX_ERROR_DTO_PAMAMS)
async def test_helpers_wrapper_raise_on_absolidix_error(x: Any, expected: bool):
    "Test raise_on_absolidix_error"

    raises = pytest.raises(AbsolidixError) if expected else nullcontext()

    with raises:
        await raise_on_absolidix_error(mk_async_fun(x))()


@pytest.mark.parametrize("x,expected", IS_ABSOLIDIX_ERRORS_EVT_DTO_PARAMS)
async def test_helpers_wrapper_raise_on_absolidix_error_in_event(
    x: Any, expected: bool
):
    "Test raise_on_absolidix_error_in_event"

    raises = pytest.raises(AbsolidixError) if expected else nullcontext()

    with raises:
        await raise_on_absolidix_error_in_event(mk_async_fun(x))()



"""Base of DTOs"""

from datetime import datetime

from ..compat import NotRequired, TypedDict


class AbsolidixTimestampsDTO(TypedDict):
    "Response with timestamps"

    created_at: NotRequired[datetime]
    updated_at: NotRequired[datetime]



"""Error DTOs"""

from typing import Union

from ..compat import TypedDict


class AbsolidixErrorMessageDTO(TypedDict):
    "Error's raw transport error payload DTO"

    message: str


class AbsolidixErrorDTO(TypedDict):
    "Error DTO"

    status: int
    error: Union[AbsolidixErrorMessageDTO, str]



"""Absolidix API synchronous client"""

import asyncio
from collections.abc import Awaitable, Callable, Sequence
from concurrent.futures import ThreadPoolExecutor
from functools import partial, wraps
from typing import Any, Literal, Optional, TypeVar, Union, cast
from warnings import warn

from aiohttp.typedefs import StrOrURL

from absolidix_client.dtos.datasource import AbsolidixDataSourceDTO

from .absolidix_async import AbsolidixAPIAsync, AbsolidixAPIKwargs
from .compat import Concatenate, ParamSpec, Unpack
from .exc import AbsolidixAsyncRuntimeWarning
from .models.base import AbsolidixBase
from .namespaces.v0_calculations import AbsolidixCalculationOnProgressT
from .namespaces.v0_collections import AbsolidixCollectionsCreateKwargs

AsyncClientGetter = Callable[[], AbsolidixAPIAsync]
ReturnT_co = TypeVar("ReturnT_co", covariant=True)
ParamT = ParamSpec("ParamT")
TimeoutType = Union[float, Literal[False], None]


class AbsolidixNamespaceSyncBase:
    "Base for synchronous namespaces"

    _client_getter: AsyncClientGetter
    _default_timeout: TimeoutType

    def __init__(
        self, client_getter: AsyncClientGetter, default_timeout: TimeoutType = False
    ):
        self._client_getter = client_getter
        self._default_timeout = default_timeout

    def _get_timeout(self, timeout: TimeoutType) -> Optional[float]:
        "Normalize timeout value"
        if timeout is False:
            return None
        if timeout is not None:
            return timeout
        if not self._default_timeout:
            return None
        return self._default_timeout


def to_sync_with_absolidix_client(
    func: Callable[Concatenate[Any, AbsolidixAPIAsync, ParamT], Awaitable[ReturnT_co]],
) -> Callable[Concatenate[Any, ParamT], ReturnT_co]:
    """
    Wraps async AbsolidixNamespaceSync's method.
    - start AbsolidixAPIAsync client,
    - pass it to the method as first argument after Self,
    - close client
    - wrap all with async_to_sync converter
    """

    @wraps(func)
    async def inner(
        self: AbsolidixNamespaceSyncBase, *args: ParamT.args, **kwargs: ParamT.kwargs
    ) -> ReturnT_co:
        """
        Call method with timeout with self, client and other args.
        """
        timeout = self._get_timeout(cast(TimeoutType, kwargs.get("timeout", None)))  # pyright: ignore [reportPrivateUsage]
        async with self._client_getter() as client:  # pyright: ignore [reportPrivateUsage]
            return await asyncio.wait_for(func(self, client, *args, **kwargs), timeout)

    @wraps(func)
    def outer(
        self: AbsolidixNamespaceSyncBase, *args: ParamT.args, **kwargs: ParamT.kwargs
    ) -> ReturnT_co:
        """
        Execute the async method synchronously in sync and async runtime.
        """
        coro = inner(self, *args, **kwargs)
        try:
            asyncio.get_running_loop()  # Triggers RuntimeError if no running event loop

            warn(
                AbsolidixAsyncRuntimeWarning(
                    "Using a synchronous API in an asynchronous runtime. "
                    "Consider switching to AbsolidixAPIAsync."
                )
            )

            # Create a separate thread so we can block before returning
            with ThreadPoolExecutor(1) as pool:
                return pool.submit(lambda: asyncio.run(coro)).result()
        except RuntimeError:
            return asyncio.run(coro)

    return outer


class AbsolidixCalculationsNamespaceSync(AbsolidixNamespaceSyncBase):
    """Calculations endpoints namespace"""

    @to_sync_with_absolidix_client
    async def supported(self, client: AbsolidixAPIAsync):
        "Get supported calculation engines"
        return await client.calculations.supported()


class AbsolidixV0AuthNamespaceSync(AbsolidixNamespaceSyncBase):
    """Authentication endpoints namespace"""

    @to_sync_with_absolidix_client
    async def login(
        self,
        client: AbsolidixAPIAsync,
        email: str,
        password: str,
        timeout: TimeoutType = None,
    ):
        "Login"
        return await client.v0.auth.login(email, password)

    @to_sync_with_absolidix_client
    async def whoami(self, client: AbsolidixAPIAsync, timeout: TimeoutType = None):
        "Get self info"
        return await client.v0.auth.whoami()


class AbsolidixV0DatasourcesNamespaceSync(AbsolidixNamespaceSyncBase):
    """Datasources endpoints namespace"""

    @to_sync_with_absolidix_client
    async def create(
        self,
        client: AbsolidixAPIAsync,
        content: str,
        fmt: Optional[str] = None,
        name: Optional[str] = None,
        timeout: TimeoutType = None,
    ):
        "Create data source and wait for the result"
        return await client.v0.datasources.create(content, fmt, name)

    @to_sync_with_absolidix_client
    async def delete(
        self, client: AbsolidixAPIAsync, data_id: int, timeout: TimeoutType = None
    ):
        "Delete data source by id and wait for the result"
        return await client.v0.datasources.delete(data_id)

    @to_sync_with_absolidix_client
    async def list(self, client: AbsolidixAPIAsync, timeout: TimeoutType = None):
        "List data sources and wait for the result"
        return await client.v0.datasources.list()

    @to_sync_with_absolidix_client
    async def get(
        self, client: AbsolidixAPIAsync, data_id: int, timeout: TimeoutType = None
    ) -> Optional[AbsolidixDataSourceDTO]:
        "Get data source by id"
        return await client.v0.datasources.get(data_id)

    @to_sync_with_absolidix_client
    async def get_parents(
        self, client: AbsolidixAPIAsync, data_id: int, timeout: TimeoutType = None
    ) -> Sequence[AbsolidixDataSourceDTO]:
        "Get parent data sources by id"
        return await client.v0.datasources.get_parents(data_id)

    @to_sync_with_absolidix_client
    async def get_children(
        self, client: AbsolidixAPIAsync, data_id: int, timeout: TimeoutType = None
    ) -> Sequence[AbsolidixDataSourceDTO]:
        "Get children data sources by id"
        return await client.v0.datasources.get_children(data_id)

    @to_sync_with_absolidix_client
    async def get_content(
        self, client: AbsolidixAPIAsync, data_id: int, timeout: TimeoutType = None
    ):
        "Get data source by id"
        return await client.v0.datasources.get_content(data_id)


class AbsolidixV0CalculationsNamespaceSync(AbsolidixNamespaceSyncBase):
    """Calculations endpoints namespace"""

    @to_sync_with_absolidix_client
    async def cancel(
        self, client: AbsolidixAPIAsync, calc_id: int, timeout: TimeoutType = None
    ):
        "Cancel calculation and wait for the result"
        return await client.v0.calculations.cancel(calc_id)

    @to_sync_with_absolidix_client
    async def create(
        self,
        client: AbsolidixAPIAsync,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,
        timeout: TimeoutType = None,
    ):
        "Create calculation and wait for the result"
        return await client.v0.calculations.create(data_id, engine, input)

    @to_sync_with_absolidix_client
    async def get_results(
        self,
        client: AbsolidixAPIAsync,
        calc_id: int,
        on_progress: Optional[AbsolidixCalculationOnProgressT] = None,
        timeout: TimeoutType = None,
    ):
        "Waits for the end of the calculation and returns the results"
        return await client.v0.calculations.get_results(calc_id, on_progress)

    @to_sync_with_absolidix_client
    async def create_get_results(
        self,
        client: AbsolidixAPIAsync,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,
        on_progress: Optional[AbsolidixCalculationOnProgressT] = None,
        timeout: TimeoutType = None,
    ):
        "Create calculation, wait done and get results"
        return await client.v0.calculations.create_get_results(
            data_id, engine, input, on_progress
        )

    @to_sync_with_absolidix_client
    async def get_engines(self, client: AbsolidixAPIAsync, timeout: TimeoutType = None):
        "Get supported calculation engines"
        return await client.v0.calculations.get_engines()

    @to_sync_with_absolidix_client
    async def list(self, client: AbsolidixAPIAsync, timeout: TimeoutType = None):
        "List all user's calculations and wait for the result"
        return await client.v0.calculations.list()

    @to_sync_with_absolidix_client
    async def get(
        self, client: AbsolidixAPIAsync, calc_id: int, timeout: TimeoutType = None
    ):
        "Get calculation by id"
        return await client.v0.calculations.get(calc_id)


class AbsolidixV0CollectionsNamespaceSync(AbsolidixNamespaceSyncBase):
    """Collections endpoints namespace"""

    @to_sync_with_absolidix_client
    async def create(
        self,
        client: AbsolidixAPIAsync,
        type_id: int,
        title: str,
        timeout: TimeoutType = None,
        **opts: Unpack[AbsolidixCollectionsCreateKwargs],
    ):
        "Create collection and wait for the result"
        return await client.v0.collections.create(type_id, title, **opts)

    @to_sync_with_absolidix_client
    async def list(self, client: AbsolidixAPIAsync, timeout: TimeoutType = None):
        "List user's collections by criteria and wait for the result"
        return await client.v0.collections.list()

    @to_sync_with_absolidix_client
    async def delete(
        self, client: AbsolidixAPIAsync, collection_id: int, timeout: TimeoutType = None
    ):
        "Remove a collection by id and wait for the result"
        return await client.v0.collections.delete(collection_id)


class AbsolidixV0NamespaceSync(AbsolidixNamespaceSyncBase):
    """v0 namespace"""

    def __init__(
        self, client_getter: AsyncClientGetter, default_timeout: Optional[float] = False
    ):
        super().__init__(client_getter, default_timeout)
        self.auth = AbsolidixV0AuthNamespaceSync(client_getter, default_timeout)
        self.calculations = AbsolidixV0CalculationsNamespaceSync(
            client_getter, default_timeout
        )
        self.collections = AbsolidixV0CollectionsNamespaceSync(
            client_getter, default_timeout
        )
        self.datasources = AbsolidixV0DatasourcesNamespaceSync(
            client_getter, default_timeout
        )


class AbsolidixAPI(AbsolidixBase):
    """Absolidix API synchronous client"""

    def __init__(
        self,
        base_url: StrOrURL,
        **opts: Unpack[AbsolidixAPIKwargs],
    ):
        """
        Initialize sync Absolidix client.
        Sync client is a tiny wrapper over full async client.

        **Arguments**:

        `bese_url`
        URL of Absolidix BFF server. `str` or `yarl.URL`.

        `headers` (Optional)
        Optional dictionary of always sent headers. Used if `session` is omitted.

        `timeout` (Optional)
        Optional float timeout in seconds. None if no timeout. False if not set.
        Used as `aiohttp` timeout if `session` is omitted.
        Used as global timeout for SSE responses.

        `client_name` (Optional)
        Optional string for user agent.
        """
        timeout = opts.get("timeout", None)
        self._ns_calculations = AbsolidixCalculationsNamespaceSync(
            partial(AbsolidixAPIAsync, base_url, **opts), timeout
        )
        self._ns_v0 = AbsolidixV0NamespaceSync(
            partial(AbsolidixAPIAsync, base_url, **opts), timeout
        )

    @property
    def calculations(
        self,
    ) -> AbsolidixCalculationsNamespaceSync:
        """Property to access the calculations namespace."""
        return self._ns_calculations

    @property
    def v0(self) -> AbsolidixV0NamespaceSync:
        """Property to access the v0 namespace."""
        return self._ns_v0



"""All data transfer objects"""

from .auth import AbsolidixAuthCredentialsRequestDTO
from .base import AbsolidixTimestampsDTO
from .calculation import AbsolidixCalculationDTO
from .collection import (
    AbsolidixCollectionCreateDTO,
    AbsolidixCollectionDTO,
    AbsolidixCollectionTypeDTO,
    AbsolidixCollectionVisibility,
)
from .datasource import (
    AbsolidixDataSourceContentOnlyDTO,
    AbsolidixDataSourceDTO,
    DataSourceType,
)
from .error import AbsolidixErrorDTO, AbsolidixErrorMessageDTO
from .event import (
    AbsolidixCalculationsEventDTO,
    AbsolidixCollectionsEventDTO,
    AbsolidixDataSourcesEventDTO,
    AbsolidixErrorEventDataDTO,
    AbsolidixErrorEventDTO,
    AbsolidixEventDTO,
    AbsolidixEventType,
    AbsolidixPongEventDTO,
)
from .resp import AbsolidixRequestIdDTO
from .user import AbsolidixUserDTO

__all__ = [
    "AbsolidixAuthCredentialsRequestDTO",
    "AbsolidixCollectionCreateDTO",
    "AbsolidixCollectionDTO",
    "AbsolidixCollectionTypeDTO",
    "AbsolidixCollectionVisibility",
    "AbsolidixCalculationDTO",
    "AbsolidixDataSourceContentOnlyDTO",
    "AbsolidixDataSourceDTO",
    "DataSourceType",
    "AbsolidixErrorDTO",
    "AbsolidixErrorMessageDTO",
    "AbsolidixEventDTO",
    "AbsolidixEventType",
    "AbsolidixPongEventDTO",
    "AbsolidixRequestIdDTO",
    "AbsolidixTimestampsDTO",
    "AbsolidixUserDTO",
    "AbsolidixCalculationsEventDTO",
    "AbsolidixCollectionsEventDTO",
    "AbsolidixDataSourcesEventDTO",
    "AbsolidixErrorEventDataDTO",
    "AbsolidixErrorEventDTO",
]



"""SSE DTOs"""

from typing import Literal, Union
from collections.abc import Sequence
from ..compat import TypedDict
from .calculation import AbsolidixCalculationDTO
from .collection import AbsolidixCollectionDTO, AbsolidixCollectionTypeDTO
from .datasource import AbsolidixDataSourceDTO
from .error import AbsolidixErrorDTO

AbsolidixEventType = Literal[
    "calculations", "collections", "datasources", "errors", "pong"
]


class AbsolidixErrorEventDataDTO(TypedDict):
    "Errors event data DTO"

    req_id: str
    data: Sequence[AbsolidixErrorDTO]


class AbsolidixErrorEventDTO(TypedDict):
    "Errors event DTO"

    type: Literal["errors"]
    data: AbsolidixErrorEventDataDTO


class AbsolidixPongEventDTO(TypedDict):
    "Pong event DTO"

    type: Literal["ping"]
    data: Literal["pong"]


class AbsolidixDataSourcesEventDataDTO(TypedDict):
    "Data sources event data DTO"

    req_id: str
    data: Sequence[AbsolidixDataSourceDTO]
    total: int
    types: Sequence[AbsolidixCollectionTypeDTO]


class AbsolidixDataSourcesEventDTO(TypedDict):
    "Data sources event DTO"

    type: Literal["datasources"]
    data: AbsolidixDataSourcesEventDataDTO


class AbsolidixCalculationsEventDataDTO(TypedDict):
    "Calculations event data DTO"

    req_id: str
    data: Sequence[AbsolidixCalculationDTO]
    total: int
    types: Sequence[AbsolidixCollectionTypeDTO]


class AbsolidixCalculationsEventDTO(TypedDict):
    "Calculations event DTO"

    type: Literal["calculations"]
    data: AbsolidixCalculationsEventDataDTO


class AbsolidixCollectionsEventDataDTO(TypedDict):
    "Collections event data DTO"

    req_id: str
    data: Sequence[AbsolidixCollectionDTO]
    total: int
    types: Sequence[AbsolidixCollectionTypeDTO]


class AbsolidixCollectionsEventDTO(TypedDict):
    "Collections event DTO"

    type: Literal["collections"]
    data: AbsolidixCollectionsEventDataDTO


AbsolidixEventDTO = Union[
    AbsolidixCalculationsEventDTO,
    AbsolidixCollectionsEventDTO,
    AbsolidixDataSourcesEventDTO,
    AbsolidixErrorEventDTO,
    AbsolidixPongEventDTO,
]



"""Custom exceptions for absolidix_client."""

from typing import Optional


class AbsolidixException(BaseException):
    """
    This is raised when unknown exceptions occour.
    And it's used as a base for all other exceptions
    so if you want to catch all Absolidix related errors
    you should catch this base exception.
    """


class AbsolidixConnectionException(AbsolidixException):
    """This is raised when there is a connection issue with Absolidix."""


class AbsolidixError(AbsolidixException):
    """Base of all errors based on results"""

    status: int = 0
    message: Optional[str]

    def __init__(
        self,
        *args: object,
        status: Optional[int],
        message: Optional[str] = None,
    ):
        super().__init__(*args)
        if status:
            self.status = status
        if message:
            self.message = message

    def __str__(self) -> str:
        return f"status: {self.status}, message: {self.message}"


class AbsolidixNotFoundException(AbsolidixError):
    """This is raised when the requested resource is not found."""


class AbsolidixPayloadException(AbsolidixError):
    """This is raised when the payload is invalid."""


class AbsolidixPermissionException(AbsolidixError):
    """This is raised when the user has no permission to do the requested resource."""


class AbsolidixAuthenticationException(AbsolidixError):
    """This is raised when we recieve an authentication issue."""


class AbsolidixQuotaException(AbsolidixError):
    """This is raised when quota excided."""


class AbsolidixAsyncRuntimeWarning(UserWarning):
    """Use of sync API in async runtime"""



"""Collections DTOs"""

from collections.abc import Sequence
from typing import Literal, Union

from ..compat import NotRequired, TypedDict
from .base import AbsolidixTimestampsDTO

AbsolidixCollectionVisibility = Union[
    Literal["private"], Literal["shared"], Literal["community"]
]


class AbsolidixCollectionTypeDTO(AbsolidixTimestampsDTO):
    "Collection type DTO"

    id: int
    slug: str
    label: str
    flavor: str


class AbsolidixCollectionCommonDTO(TypedDict):
    "Common fields of collection DTOs"

    title: str
    type_id: int
    datasources: NotRequired[Sequence[int]]
    users: NotRequired[Sequence[int]]


class AbsolidixCollectionCreateDTO(AbsolidixCollectionCommonDTO):
    "Collection create DTO"

    id: NotRequired[int]
    description: NotRequired[str]
    visibility: NotRequired[AbsolidixCollectionVisibility]


class AbsolidixCollectionDTO(AbsolidixCollectionCommonDTO, AbsolidixTimestampsDTO):
    "Collection DTO"

    id: int
    description: str
    visibility: AbsolidixCollectionVisibility

    user_id: int
    user_first_name: NotRequired[str]
    user_last_name: NotRequired[str]
    type_slug: NotRequired[str]
    type_label: NotRequired[str]
    type_flavor: NotRequired[str]



"""Calculations endpoints namespace"""

from collections.abc import Awaitable, Sequence
from datetime import datetime
from functools import partial
from inspect import iscoroutinefunction
from typing import Callable, Optional, Union, cast
from warnings import warn

from ..dtos import (
    AbsolidixCalculationDTO,
    AbsolidixDataSourceDTO,
    AbsolidixRequestIdDTO,
    DataSourceType,
)
from ..exc import AbsolidixPayloadException
from ..helpers import absolidix_json_decoder, raise_on_absolidix_error
from ..models import act_and_get_result_from_stream
from .base import BaseNamespace

DATA_SOURCE_CALC_RESULT_TYPES = [DataSourceType.PROPERTY, DataSourceType.PATTERN]

AbsolidixCalculationOnProgressT = Callable[
    [AbsolidixCalculationDTO], Union[Optional[bool], Awaitable[Optional[bool]]]
]


class AbsolidixV0CalculationsNamespace(BaseNamespace):
    """Calculations endpoints namespace"""

    async def cancel_event(self, calc_id: int) -> AbsolidixRequestIdDTO:
        "Cancel calculation"
        async with await self._client.request(
            method="DELETE",
            url=self._base_url / str(calc_id),
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)

    async def cancel(self, calc_id: int) -> None:
        "Cancel calculation and wait for result"
        await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.cancel_event, calc_id)
        )

    async def create_event(
        self,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,
    ) -> AbsolidixRequestIdDTO:
        "Create calculation"
        async with await self._client.request(
            method="POST",
            url=self._base_url,
            json={"dataId": data_id, "engine": engine, "input": input},
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)

    async def create(
        self,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,
    ) -> Optional[AbsolidixCalculationDTO]:
        "Create calculation and wait for result"
        valid_engines = await self._root.calculations.supported()
        if engine not in valid_engines:
            raise AbsolidixPayloadException(message="unsupported engine", status=400)
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe,
            partial(self.create_event, data_id, engine, input),
        )
        if evt["type"] == "calculations":
            data = sorted(
                evt.get("data", {}).get("data", []),
                key=lambda x: x.get("created_at", datetime.fromordinal(1)),
            )
            return data[-1] if data else None

    async def _get_results(
        self,
        calc_getter: Callable[[], Awaitable[Optional[AbsolidixCalculationDTO]]],
        on_progress: Optional[AbsolidixCalculationOnProgressT] = None,
    ) -> Optional[Sequence[AbsolidixDataSourceDTO]]:
        "Waits for the end of the calculation and returns the results"

        def get_new_calc_id(data_id: int, calcs: Sequence[AbsolidixCalculationDTO]):
            """
            The funny part is that when the computation ends,
            we get a data source instead of a calculation.
            """
            for calc in calcs:
                if calc.get("progress", 0) < 100:
                    continue
                data = cast(AbsolidixDataSourceDTO, calc)
                if data_id in data.get("parents", []):
                    return data["id"]
            return None

        def get_calc_from_listing(
            calc_id: int, calcs: Sequence[AbsolidixCalculationDTO]
        ):
            for calc in calcs:
                if calc_id == calc["id"]:
                    return calc
            return None

        def filter_ds_for_calc(data_id: int, dss: Sequence[AbsolidixDataSourceDTO]):
            return [
                ds
                for ds in dss
                if ds["type"] in DATA_SOURCE_CALC_RESULT_TYPES
                and data_id in ds["parents"]
            ]

        async with self._root.stream.subscribe() as sub:
            target_calc = await calc_getter()
            if not target_calc:
                return  # pragma: no cover
            calc_id = target_calc["id"]
            data_id = target_calc["parent"]
            async for msg in sub:
                if msg["type"] == "calculations":
                    calc_id = get_new_calc_id(data_id, msg["data"]["data"]) or calc_id
                    target_calc = get_calc_from_listing(calc_id, msg["data"]["data"])

                    # run callback if any and exit if needed
                    if target_calc and on_progress:
                        if (
                            await on_progress(target_calc)
                            if iscoroutinefunction(on_progress)
                            else on_progress(target_calc)
                        ) is False:
                            return

                # results
                if msg["type"] == "datasources":
                    results = filter_ds_for_calc(data_id, msg["data"]["data"])
                    # if results or calc is done but no results
                    if results or target_calc is None:
                        return results

    async def get_results(
        self,
        calc_id: int,
        on_progress: Optional[AbsolidixCalculationOnProgressT] = None,
    ) -> Optional[Sequence[AbsolidixDataSourceDTO]]:
        "Waits for the end of the calculation and returns the results"

        async def get_calc():
            return await self.get(calc_id)

        return await self._get_results(get_calc, on_progress)

    async def create_get_results(
        self,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,
        on_progress: Optional[AbsolidixCalculationOnProgressT] = None,
    ) -> Optional[Sequence[AbsolidixDataSourceDTO]]:
        "Create calculation, wait done and get results"

        async def get_calc():
            return await self.create(data_id, engine, input)

        return await self._get_results(get_calc, on_progress)

    @raise_on_absolidix_error
    async def get_engines(self) -> Sequence[str]:
        "Get supported calculation engines"
        msg = (
            "v0.calculations.get_engines() is deprecated; "
            "please, use calculations.supported() instead"
        )
        warn(DeprecationWarning(msg))
        return await self._root.calculations.supported()

    async def list_event(self) -> AbsolidixRequestIdDTO:
        "List all user's calculations"
        async with await self._client.request(
            method="GET",
            url=self._base_url,
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)

    async def list(self) -> Sequence[AbsolidixCalculationDTO]:
        "List all user's calculations and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.list_event)
        )
        if evt["type"] == "calculations":
            return evt.get("data", {}).get("data", [])
        return []  # pragma: no cover

    async def get(self, calc_id: int) -> Optional[AbsolidixCalculationDTO]:
        "Get calculation by id"
        data = list(filter(lambda x: x["id"] == calc_id, await self.list()))
        return data[-1] if data else None



"""Authenticators"""

from abc import abstractmethod
from asyncio import Lock, sleep

from aiohttp import ClientSession
from aiohttp.hdrs import METH_POST
from aiohttp.web_exceptions import HTTPTooManyRequests
from yarl import URL

from ..dtos import AbsolidixAuthCredentialsRequestDTO
from .base import AbsolidixBase


class BaseAuthenticator(AbsolidixBase):
    """Base authentication class"""

    lock: Lock

    def __init__(self):
        self.lock = Lock()

    @abstractmethod
    async def authenticate(self, session: ClientSession, base_url: URL) -> bool:
        "Run authentication procedure"

    @abstractmethod
    async def should_update(self, session: ClientSession, base_url: URL) -> bool:
        "Check if authentication needed"


class AbsolidixNoAuth(BaseAuthenticator):
    """No authentication (noop)"""

    async def authenticate(self, session: ClientSession, base_url: URL) -> bool:
        return True

    async def should_update(self, session: ClientSession, base_url: URL) -> bool:
        return False


class AbsolidixTokenAuth(BaseAuthenticator):
    """Token based authentication"""

    _token: str

    def __init__(self, token: str) -> None:
        super().__init__()
        self._token = token

    async def authenticate(self, session: ClientSession, base_url: URL) -> bool:
        session.headers["Authorization"] = f"Bearer {self._token}"
        return True

    async def should_update(self, session: ClientSession, base_url: URL) -> bool:
        return True


class AbsolidixLocalUserAuth(BaseAuthenticator):
    """Password based authentication"""

    endpoint = "v0/auth"
    _credentials: AbsolidixAuthCredentialsRequestDTO
    _cookie_name = "_sid"
    # force authentication in case when cookie already exists
    # in a session due to an unauthenticated request
    _force = True

    def __init__(self, email: str, password: str) -> None:
        super().__init__()
        self._credentials = AbsolidixAuthCredentialsRequestDTO(
            email=email, password=password
        )
        self.logger.warning(
            "Please, do NOT use password-based authentication, it is for testing only"
        )

    @classmethod
    def _get_cookie(cls, session: ClientSession, base_url: URL):
        return session.cookie_jar.filter_cookies(base_url).get(cls._cookie_name)

    async def authenticate(self, session: ClientSession, base_url: URL) -> bool:
        async with session.request(
            METH_POST,
            base_url / self.endpoint,
            json=self._credentials,
            raise_for_status=False,
        ) as resp:
            if resp.status == HTTPTooManyRequests.status_code:
                await sleep(10)
                return await self.authenticate(session, base_url)
            if not resp.ok:
                session.cookie_jar.clear(lambda x: x.key == self._cookie_name)
                return False
            self._force = False
            return bool(self._get_cookie(session, base_url))

    async def should_update(self, session: ClientSession, base_url: URL) -> bool:
        session.cookie_jar.update_cookies({}, base_url)
        return self._force or not bool(self._get_cookie(session, base_url))



"""Root namespace"""

from yarl import URL

from ..client import AbsolidixClient
from .base import BaseNamespace
from .calculations import AbsolidixCalculationsNamespace
from .stream import AbsolidixStreamNamespace
from .v0 import AbsolidixV0Namespace


class AbsolidixRootNamespace(BaseNamespace):
    """Root namespace"""

    def __init__(
        self,
        client: AbsolidixClient,
        base_url: URL,
        root: None = None,
        auth_req: bool = True,
    ) -> None:
        """Initialise the namespace."""
        super().__init__(client, base_url, root or self, auth_req)

    def __post_init__(self) -> None:
        self.__ns_calculations = AbsolidixCalculationsNamespace(
            self._client, self._base_url / "calculations", root=self
        )
        self.__ns_v0 = AbsolidixV0Namespace(
            self._client, self._base_url / "v0", root=self
        )
        self.__ns_stream = AbsolidixStreamNamespace(
            self._client, self._base_url / "stream", root=self
        )

    @property
    def calculations(self) -> "AbsolidixCalculationsNamespace":
        """Property to access the calculations namespace."""
        return self.__ns_calculations

    @property
    def v0(self) -> "AbsolidixV0Namespace":
        """Property to access the v0 namespace."""
        return self.__ns_v0

    @property
    def stream(self) -> "AbsolidixStreamNamespace":
        """Property to access the stream namespace."""
        return self.__ns_stream



"""Stream endpoint namespace"""

import asyncio
from collections.abc import Callable
from typing import NoReturn, Optional

from aiohttp_sse_client.client import MessageEvent

from ..dtos import AbsolidixEventDTO
from ..models import AbsolidixHub, AbsolidixMessageEvent, AbsolidixSubscription
from .base import BaseNamespace


class AbsolidixStreamNamespace(BaseNamespace):
    """Stream endpoints namespace"""

    _hub: AbsolidixHub
    _stream_task: Optional[asyncio.Task[NoReturn]] = None
    _sse_client_task: Optional[asyncio.Task[None]] = None
    _subscribe_event: asyncio.Event

    def __post_init__(self) -> None:
        self._hub = AbsolidixHub()
        self._stream_task = asyncio.create_task(
            self._stream_consumer(), name="StreamConsumerTask"
        )
        self._subscribe_event = asyncio.Event()
        return super().__post_init__()

    async def _stream_consumer(self):
        def on_open():
            self._hub.set_connected()

        def on_message(evt: MessageEvent):
            self._hub.set_connected()
            evt_dto = AbsolidixMessageEvent.from_dto(evt).to_dto()
            self._hub.publish(evt_dto)
            # cancel streaming task if no subscribers
            if self._sse_client_task and len(self._hub) == 0:
                self._hub.set_disconnected()
                self._sse_client_task.cancel()

        while True:
            await self._subscribe_event.wait()
            if self._sse_client_task is None or self._sse_client_task.done():
                self._sse_client_task = asyncio.create_task(
                    self._client.sse(self._base_url, on_message, on_open),
                    name="SSEClientTask",
                )
            self._subscribe_event.clear()
            while not self._hub.connected:
                await asyncio.sleep(0.1)
                await self._root.v0.ping()

    def close(self):
        "Close background stream consumer"
        if self._stream_task:
            self._stream_task.cancel()
        if self._sse_client_task:
            self._sse_client_task.cancel()

    def subscribe(
        self, predicate: Optional[Callable[[AbsolidixEventDTO], bool]] = None
    ):
        "Subscribe to stream"
        self._subscribe_event.set()
        return AbsolidixSubscription(self._hub, predicate=predicate)



"Test type guards"

from typing import Any, Union

import pytest

from absolidix_client.dtos import (
    AbsolidixErrorDTO,
    AbsolidixErrorEventDataDTO,
    AbsolidixErrorEventDTO,
    AbsolidixErrorMessageDTO,
)
from absolidix_client.helpers import (
    is_absolidix_error_dto,
    is_absolidix_error_error_dto,
    is_absolidix_error_event_data_dto,
    is_absolidix_errors_evt_dto,
)

GOOD_ERROR_MSG: AbsolidixErrorMessageDTO = {"message": "message"}
BAD_ERROR_MSG1 = "error error"
BAD_ERROR_MSG2: object = {}
BAD_ERROR_MSG3 = {"message": None}


@pytest.mark.parametrize(
    "x,expected",
    [
        (GOOD_ERROR_MSG, True),
        (BAD_ERROR_MSG1, False),
        (BAD_ERROR_MSG2, False),
        (BAD_ERROR_MSG3, False),
    ],
)
def test_is_absolidix_error_error_dto(
    x: Union[AbsolidixErrorMessageDTO, str, None], expected: bool
):
    "Test is_absolidix_error_error_dto"
    assert is_absolidix_error_error_dto(x) == expected


GOOD_ERROR1: AbsolidixErrorDTO = {"status": 1, "error": "error message"}
GOOD_ERROR2: AbsolidixErrorDTO = {"status": 1, "error": GOOD_ERROR_MSG}
BAD_ERROR1 = "error"
BAD_ERROR2 = {"status": 1}
BAD_ERROR3 = {"error": GOOD_ERROR_MSG}
BAD_ERROR4 = {"status": 1, "error": BAD_ERROR_MSG2}
IS_ABSOLIDIX_ERROR_DTO_PAMAMS = [
    (GOOD_ERROR1, True),
    (GOOD_ERROR2, True),
    (BAD_ERROR1, False),
    (BAD_ERROR2, False),
    (BAD_ERROR3, False),
    (BAD_ERROR4, False),
]


@pytest.mark.parametrize("x,expected", IS_ABSOLIDIX_ERROR_DTO_PAMAMS)
def test_is_absolidix_error_dto(x: Union[AbsolidixErrorDTO, str, None], expected: bool):
    "Test is_absolidix_error_dto"
    assert is_absolidix_error_dto(x) == expected


GOOD_ERROR_EVENT_DATA1: AbsolidixErrorEventDataDTO = {
    "req_id": "id",
    "data": [GOOD_ERROR1, GOOD_ERROR2],
}
GOOD_ERROR_EVENT_DATA2: AbsolidixErrorEventDataDTO = {"req_id": "id", "data": []}
BAD_ERROR_EVENT_DATA1 = "error event data"
BAD_ERROR_EVENT_DATA2: dict[str, Any] = {}
BAD_ERROR_EVENT_DATA3 = {"req_id": "id"}
BAD_ERROR_EVENT_DATA4: dict[str, list[None]] = {"data": []}
BAD_ERROR_EVENT_DATA5 = {"req_id": "id", "data": ""}
BAD_ERROR_EVENT_DATA6 = {"req_id": "id", "data": [BAD_ERROR1]}
BAD_ERROR_EVENT_DATA7 = {"req_id": "id", "data": [GOOD_ERROR1, BAD_ERROR1]}


@pytest.mark.parametrize(
    "x,expected",
    [
        (GOOD_ERROR_EVENT_DATA1, True),
        (GOOD_ERROR_EVENT_DATA2, True),
        (BAD_ERROR_EVENT_DATA1, False),
        (BAD_ERROR_EVENT_DATA2, False),
        (BAD_ERROR_EVENT_DATA3, False),
        (BAD_ERROR_EVENT_DATA4, False),
        (BAD_ERROR_EVENT_DATA5, False),
        (BAD_ERROR_EVENT_DATA6, False),
        (BAD_ERROR_EVENT_DATA7, False),
    ],
)
def test_is_absolidix_error_event_data_dto(x: Any, expected: bool):
    "Test is_absolidix_error_event_data_dto"
    assert is_absolidix_error_event_data_dto(x) == expected


GOOD_ERROR_EVENT1: AbsolidixErrorEventDTO = {
    "type": "errors",
    "data": GOOD_ERROR_EVENT_DATA1,
}
BAD_ERROR_EVENT1 = "event"
BAD_ERROR_EVENT2 = {"type": "no"}
BAD_ERROR_EVENT3 = {"type": "errors"}
BAD_ERROR_EVENT4 = {"type": "errors", "data": "data"}
IS_ABSOLIDIX_ERRORS_EVT_DTO_PARAMS = [
    (GOOD_ERROR_EVENT1, True),
    (BAD_ERROR_EVENT1, False),
    (BAD_ERROR_EVENT2, False),
    (BAD_ERROR_EVENT3, False),
    (BAD_ERROR_EVENT4, False),
]


@pytest.mark.parametrize("x,expected", IS_ABSOLIDIX_ERRORS_EVT_DTO_PARAMS)
def test_is_absolidix_errors_evt_dto(x: Any, expected: bool):
    "Test is_absolidix_errors_evt_dto"
    assert is_absolidix_errors_evt_dto(x) == expected



"Test helper encoders and decoders"

import json
from datetime import datetime
from typing import Any

import pytest

from absolidix_client.helpers import (
    absolidix_json_decoder,
    absolidix_json_encoder,
    convert_dict_values_from_dt,
    convert_dict_values_to_dt,
    parse_rfc3339,
)

dt_testdata = [
    ({"some": 123}, {"some": 123}),
    (
        {"date": "2001-02-03T04:05:06", "str": "str", "int": int},
        {"date": parse_rfc3339("2001-02-03 04:05:06"), "str": "str", "int": int},
    ),
    (
        {"date": "2001-02-03T04:05:06", "other_date": "2001-02-03T04:05:06+00:00"},
        {
            "date": parse_rfc3339("2001-02-03 04:05:06"),
            "other_date": parse_rfc3339("2001-02-03T04:05:06+00:00"),
        },
    ),
    (
        {"date": "2024-02-02T16:54:11.788905"},
        {"date": parse_rfc3339("2024-02-02T16:54:11.788905")},
    ),
    ({"date": ""}, {"date": ""}),
    ({"date": None}, {"date": None}),
    (
        {"children": [{"date": "2001-02-03T04:05:06"}]},
        {"children": [{"date": parse_rfc3339("2001-02-03 04:05:06")}]},
    ),
    (
        {"children": ({"date": "2001-02-03T04:05:06"},)},
        {"children": ({"date": parse_rfc3339("2001-02-03 04:05:06")},)},
    ),
]

dt_testdata_to_dt = [
    *dt_testdata,
    ({"date": "2001-02-03 04:05:06"}, {"date": parse_rfc3339("2001-02-03 04:05:06")}),
    ({"date": "2001-02-03T04:05:06Z"}, {"date": parse_rfc3339("2001-02-03 04:05:06Z")}),
]


@pytest.mark.parametrize("x,expected", dt_testdata_to_dt)
def test_convert_to_dt(x: dict[str, Any], expected: dict[str, Any]):
    "Test convert_dict_values_to_dt()"
    y = convert_dict_values_to_dt(x)
    assert y == expected, "Converted should match expected value"


@pytest.mark.parametrize("expected,x", dt_testdata)
def test_convert_from_dt(x: dict[str, Any], expected: dict[str, Any]):
    "Test convert_dict_values_to_dt()"
    y = convert_dict_values_from_dt(x)
    assert y == expected, "Converted should match expected value"


dt_now = datetime.now()
json_testdata = [
    ({"test_id": 1}, json.dumps({"test_id": 1})),
    ({"date_time": dt_now}, json.dumps({"date_time": dt_now.isoformat()})),
    ({"nested_dict": {"test_id": 1}}, json.dumps({"nested_dict": {"test_id": 1}})),
    ({"nested_list": [{"test_id": 1}]}, json.dumps({"nested_list": [{"test_id": 1}]})),
]


@pytest.mark.parametrize("x,expected", json_testdata)
def test_json_encoder(x: dict[str, Any], expected: dict[str, Any]):
    "Test absolidix_json_encoder()"
    y = absolidix_json_encoder(x)
    assert y == expected, "Converted should match expected value"


@pytest.mark.parametrize("expected,x", json_testdata)
def test_json_decoder(x: str, expected: dict[str, Any]):
    "Test absolidix_json_decoder()"
    y = absolidix_json_decoder(x)
    assert y == expected, "Converted should match expected value"



"Test AbsolidixClient"

import asyncio
import json
import logging
from asyncio import Task
from contextlib import suppress
from itertools import count
from typing import Optional, Union

import freezegun
import pytest
from aiohttp import web
from aiohttp.test_utils import TestServer
from aiohttp.web_exceptions import (
    HTTPBadRequest,
    HTTPException,
    HTTPForbidden,
    HTTPInternalServerError,
    HTTPMisdirectedRequest,
    HTTPNotFound,
    HTTPOk,
    HTTPPaymentRequired,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from aiohttp_sse_client.client import MessageEvent
from yarl import URL

from absolidix_client import AbsolidixNoAuth, AbsolidixTokenAuth
from absolidix_client.client import AbsolidixClient
from absolidix_client.exc import (
    AbsolidixAuthenticationException,
    AbsolidixConnectionException,
    AbsolidixError,
    AbsolidixException,
    AbsolidixNotFoundException,
    AbsolidixPayloadException,
    AbsolidixQuotaException,
)
from tests.helpers import TestClient, random_word

freezegun.configure(extend_ignore_list=["aiohttp"])

TOKEN = random_word(10)

PATH_CHECK_TOKEN_AUTH = "/check_bearer"
PATH_CHECK_PARAMS = "/check_params"
PATH_CHECK_HEADER = "/check_header"
PATH_CHECK_DATA = "/check_data"
PATH_TOO_MANY = "/too_many"
TOO_MANY_COUNTER = count(0)
PATH_SLOW_RESPONSE = "/slow"
PATH_PAYLOAD_ERROR = "/payload_error"
PATH_ECHO_STATUS = "/echo_status"
PATH_JSON_CASE_CHECK_STATUS = "/json_case_check"
PATH_SSE_SIMPLE = "/sse"


async def check_token_auth_handler(request: web.Request) -> web.Response:
    "Request handler"
    if request.headers.get("Authorization") == f"Bearer {TOKEN}":
        return web.Response(status=HTTPOk.status_code)
    return web.Response(status=HTTPUnauthorized.status_code)


async def check_params_handler(request: web.Request) -> web.Response:
    "Request handler"
    if request.query.get("param") == "ok":
        return web.Response(status=HTTPOk.status_code)
    return web.Response(status=HTTPBadRequest.status_code)


async def check_query_handler(request: web.Request) -> web.Response:
    "Request handler"
    if request.query.get("param") == "ok":
        return web.Response(status=HTTPOk.status_code)
    return web.Response(status=HTTPBadRequest.status_code)


async def check_header_handler(request: web.Request) -> web.Response:
    "Request handler"
    if request.headers.get("header") == "ok":
        return web.Response(status=HTTPOk.status_code)
    return web.Response(status=HTTPBadRequest.status_code)


async def check_data_handler(request: web.Request) -> web.Response:
    "Request handler"
    if await request.content.read() == b"ok":
        return web.Response(status=HTTPOk.status_code)
    return web.Response(status=HTTPBadRequest.status_code)


async def too_many_handler(_: web.Request) -> web.Response:
    "Request handler"
    if next(TOO_MANY_COUNTER) < 2:
        return web.Response(status=HTTPTooManyRequests.status_code)
    return web.Response(status=HTTPOk.status_code)


async def slow_response_handler(_: web.Request) -> web.Response:
    "Request handler"
    try:
        await asyncio.sleep(3600)
    except asyncio.CancelledError:
        pass
    return web.Response(status=HTTPOk.status_code)


async def payload_error_handler(_: web.Request) -> web.Response:
    "Request handler"
    resp = web.Response(text="text")
    resp.headers["Content-Encoding"] = "gzip"
    return resp


async def echo_status_handler(request: web.Request) -> web.Response:
    "Request handler"
    status_code = int(request.query.get("status_code", 400))
    content_type = request.query.get("content_type")
    charset = content_type and "utf-8"
    body = request.query.get("body")
    return web.Response(
        status=status_code, content_type=content_type, charset=charset, body=body
    )


async def json_case_check_status_handler(request: web.Request) -> web.Response:
    "Request handler"
    payload = await request.json()
    out_data = {"original": json.dumps(payload), "payload": payload}
    return web.json_response(out_data)


async def sse_simple_handler(request: web.Request) -> web.Response:
    "Request handler"
    status_code = int(request.query.get("force_status", HTTPOk.status_code))
    if request.headers.get("Authorization") != f"Bearer {TOKEN}" or status_code == 401:
        return web.Response(
            status=HTTPUnauthorized.status_code, content_type="text/plain"
        )

    event_type = request.query.get("event_type", "message")
    event_data = request.query.get("event_data", random_word(10))
    body = f"event: {event_type}\ndata: {event_data}\n\n"
    resp = web.Response(status=status_code, body=body)
    resp.content_type = "text/event-stream"
    resp.headers["Cache-Control"] = "no-store"
    return resp


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    app.router.add_get(PATH_CHECK_TOKEN_AUTH, check_token_auth_handler)
    app.router.add_get(PATH_CHECK_PARAMS, check_params_handler)
    app.router.add_get(PATH_CHECK_HEADER, check_header_handler)
    app.router.add_post(PATH_CHECK_DATA, check_data_handler)
    app.router.add_get(PATH_TOO_MANY, too_many_handler)
    app.router.add_get(PATH_SLOW_RESPONSE, slow_response_handler)
    app.router.add_get(PATH_PAYLOAD_ERROR, payload_error_handler)
    app.router.add_get(PATH_ECHO_STATUS, echo_status_handler)
    app.router.add_post(PATH_JSON_CASE_CHECK_STATUS, json_case_check_status_handler)
    app.router.add_get(PATH_SSE_SIMPLE, sse_simple_handler)
    return app


@pytest.fixture
async def cli():
    "Create test client"
    async with TestClient(TestServer(await create_app())) as client:
        yield client


@pytest.fixture
async def client(cli: TestClient) -> AbsolidixClient:
    "Create AbsolidixClient"
    return AbsolidixClient(
        session=cli.session,
        base_url=cli.make_url(""),
        auth=AbsolidixTokenAuth(TOKEN),
    )


async def test_relative_url(cli: TestClient):
    "Test error on relative base url"

    with pytest.raises(TypeError) as exc_info:
        AbsolidixClient(
            session=cli.session, base_url=URL("/relative"), auth=AbsolidixNoAuth()
        )
    assert "URL" in exc_info.value.args[0], (
        "AbsolidixClient should reject relative base url"
    )


async def test_token_auth_forced(
    client: AbsolidixClient,
):
    "Test successful authentication"
    resp = await client.request(URL(PATH_CHECK_TOKEN_AUTH), auth_required=True)
    assert resp.ok, "Client should authenticate"


async def test_params(
    client: AbsolidixClient,
):
    "Test query param passthrough"
    resp = await client.request(URL(PATH_CHECK_PARAMS), params={"param": "ok"})
    assert resp.ok, "Client should pass query param"


async def test_header(
    client: AbsolidixClient,
):
    "Test header passthrough"
    resp = await client.request(URL(PATH_CHECK_HEADER), headers={"header": "ok"})
    assert resp.ok, "Client should pass header"


async def test_data(
    client: AbsolidixClient,
):
    "Test data passthrough"
    resp = await client.request(URL(PATH_CHECK_DATA), data="ok", method="POST")
    assert resp.ok, "Client should pass data"


async def test_relative_target_url(
    client: AbsolidixClient,
):
    "Test relative targer url not started with '/'"
    resp = await client.request(URL(PATH_CHECK_TOKEN_AUTH[1:]))
    assert resp.ok, "Client should not fail"


@freezegun.freeze_time("1970-01-01", auto_tick_seconds=10)
async def test_too_many_request(
    client: AbsolidixClient,
):
    "Test too many requests"
    resp = await client.request(URL(PATH_TOO_MANY))
    assert resp.ok, "Should tolerate too many requests"


async def test_client_connection_error(
    client: AbsolidixClient,
):
    "Test ClientConnectionError"

    with pytest.raises(AbsolidixConnectionException) as exc_info:
        await client.request(URL("http://0.0.0.0:1"))
    assert "Request exception" in exc_info.value.args[0], (
        "AbsolidixClient should handle ClientConnectionError"
    )


async def test_client_payload_error(
    client: AbsolidixClient,
):
    "Test ClientPayloadError"

    with pytest.raises(AbsolidixException) as exc_info:
        await client.request(URL(PATH_PAYLOAD_ERROR))
    assert "Broken" in exc_info.value.args[0], (
        "AbsolidixClient should handle ClientPayloadError"
    )
    query = {
        "status_code": HTTPOk.status_code,
        "content_type": "application/json",
        "body": random_word(10),
    }
    with pytest.raises(AbsolidixException) as exc_info:
        await client.request(URL(PATH_ECHO_STATUS).with_query(query))
    assert "Broken" in exc_info.value.args[0], (
        "AbsolidixClient should handle JSONDecodeError"
    )


async def test_timeout_error(
    client: AbsolidixClient,
):
    "Test timeout"

    with pytest.raises(AbsolidixConnectionException) as exc_info:
        await client.request(URL(PATH_SLOW_RESPONSE), timeout=0.0001)
    assert "Timeout" in exc_info.value.args[0], "AbsolidixClient should handle timeout"


@pytest.mark.parametrize(
    "status,exception",
    [
        (HTTPForbidden, AbsolidixAuthenticationException),
        (HTTPUnauthorized, AbsolidixAuthenticationException),
        (HTTPNotFound, AbsolidixNotFoundException),
        (HTTPPaymentRequired, AbsolidixQuotaException),
        (HTTPMisdirectedRequest, AbsolidixError),
        (HTTPMisdirectedRequest, AbsolidixError),
        (HTTPBadRequest, AbsolidixPayloadException),
    ],
)
async def test_http_status(
    client: AbsolidixClient, status: HTTPException, exception: type[BaseException]
):
    "Test http status codes"

    def get_url(
        status: int, content_type: Optional[str] = None, body: Optional[str] = None
    ):
        query: dict[str, Union[int, str]] = {"status_code": status}
        if content_type:
            query["content_type"] = content_type
        if body:
            query["body"] = body
        return URL(PATH_ECHO_STATUS).with_query(query)

    if status == HTTPBadRequest:
        body = random_word(16)
        url = get_url(status.status_code, "text/plain", body)
        with pytest.raises(exception):
            await client.request(url)
        body = {"error": random_word(16)}
        url = get_url(
            HTTPBadRequest.status_code,
            content_type="application/json",
            body=json.dumps(body),
        )
        with pytest.raises(exception):
            await client.request(url)
    else:
        with pytest.raises(exception):
            await client.request(get_url(status.status_code))


async def test_sse_msg_receive(
    client: AbsolidixClient,
):
    "Test simple SSE event receive"

    task: Optional[Task[None]] = None
    event_type = random_word(8)
    event_data = random_word(16)

    def on_message(evt: MessageEvent):
        assert evt.type == event_type and evt.data == event_data, (
            "Should receive same SSE event as requested"
        )
        if task:
            task.cancel()

    query = {"event_type": event_type, "event_data": event_data}
    task = asyncio.create_task(
        client.sse(URL(PATH_SSE_SIMPLE).with_query(query), on_message)
    )
    await asyncio.wait_for(task, 1)


async def test_sse_timeout(
    client: AbsolidixClient,
    caplog: pytest.LogCaptureFixture,
):
    "Test reconnect on timeout"
    task: Optional[Task[None]] = None

    def on_message(_: MessageEvent):
        pass

    def timeouted_in_log() -> bool:
        recs = list(filter(lambda x: "timeouted" in x.message, caplog.records))
        return len(recs) >= 1

    async def wait_timeouted_in_log():
        while not timeouted_in_log():
            await asyncio.sleep(0.001)

    with caplog.at_level(logging.WARNING):
        task = asyncio.create_task(
            client.sse(URL(PATH_SSE_SIMPLE), on_message, timeout=0.00000001)
        )
        await asyncio.wait_for(wait_timeouted_in_log(), 1)
        task.cancel()
        await task

    assert timeouted_in_log(), "Client should log about timeout"


async def test_sse_connection_error(
    client: AbsolidixClient,
):
    "Test sse connection error"

    def on_message(_: MessageEvent):
        pass

    with pytest.raises(AbsolidixConnectionException) as exc_info:
        await asyncio.create_task(client.sse(URL("http://0.0.0.0:1"), on_message))
    assert "Connection error" in exc_info.value.args[0], (
        "Client should throw exception on connection errors"
    )


async def test_sse_response_errors(
    client: AbsolidixClient,
):
    "Test sse connection error"

    task: Optional[Task[None]] = None

    def on_message(_: MessageEvent):
        pass

    query = {"force_status": HTTPTooManyRequests.status_code}
    task = asyncio.create_task(
        client.sse(URL(PATH_SSE_SIMPLE).with_query(query), on_message)
    )
    with suppress(asyncio.TimeoutError):
        await asyncio.wait_for(task, 0.01)

    query = {"force_status": HTTPInternalServerError.status_code}
    task = asyncio.create_task(
        client.sse(URL(PATH_SSE_SIMPLE).with_query(query), on_message)
    )
    with suppress(asyncio.TimeoutError):
        await asyncio.wait_for(task, 0.01)

    query = {"force_status": HTTPNotFound.status_code}
    with pytest.raises(AbsolidixNotFoundException):
        await asyncio.create_task(
            client.sse(URL(PATH_SSE_SIMPLE).with_query(query), on_message)
        )



"""Datasources endpoints namespace"""

from collections.abc import Sequence
from datetime import datetime
from functools import partial
from typing import Optional

from ..dtos import (
    AbsolidixDataSourceContentOnlyDTO,
    AbsolidixDataSourceDTO,
    AbsolidixRequestIdDTO,
)
from ..helpers import absolidix_json_decoder, raise_on_absolidix_error
from ..models import act_and_get_result_from_stream
from .base import BaseNamespace


class AbsolidixV0DatasourcesNamespace(BaseNamespace):
    """Datasources endpoints namespace"""

    async def create_event(
        self, content: str, fmt: Optional[str] = None, name: Optional[str] = None
    ) -> AbsolidixRequestIdDTO:
        "Create data source"
        async with self._client.request(
            method="POST",
            url=self._base_url,
            json={"content": content, "fmt": fmt, "name": name},
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)

    async def create(
        self, content: str, fmt: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[AbsolidixDataSourceDTO]:
        "Create data source and wait for the result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.create_event, content, fmt, name)
        )
        if evt["type"] == "datasources":
            data = sorted(
                evt.get("data", {}).get("data", []),
                key=lambda x: x.get("created_at", datetime.fromordinal(1)),
            )
            return data[-1] if data else None

    async def delete_event(self, data_id: int) -> AbsolidixRequestIdDTO:
        "Delete data source by id"
        async with self._client.request(
            method="DELETE",
            url=self._base_url / str(data_id),
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)

    async def delete(self, data_id: int) -> None:
        "Delete data source by id and wait for the result"
        await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.delete_event, data_id)
        )

    async def list_event(self) -> AbsolidixRequestIdDTO:
        "List data sources"
        async with self._client.request(
            method="GET",
            url=self._base_url,
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)

    async def list(self) -> Sequence[AbsolidixDataSourceDTO]:
        "List data sources and wait for the result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.list_event)
        )
        if evt["type"] == "datasources":
            return evt.get("data", {}).get("data", [])
        return []  # pragma: no cover

    async def get(self, data_id: int) -> Optional[AbsolidixDataSourceDTO]:
        "Get data source by id"
        data = list(filter(lambda x: x["id"] == data_id, await self.list()))
        return data[-1] if data else None

    async def get_parents(self, data_id: int) -> Sequence[AbsolidixDataSourceDTO]:
        "Get parent data sources by id"
        return list(
            filter(lambda x: data_id in x.get("children", []), await self.list())
        )

    async def get_children(self, data_id: int) -> Sequence[AbsolidixDataSourceDTO]:
        "Get children data sources by id"
        return list(
            filter(lambda x: data_id in x.get("parents", []), await self.list())
        )

    @raise_on_absolidix_error
    async def get_content(self, data_id: int) -> AbsolidixDataSourceContentOnlyDTO:
        "Get data source by id"
        async with self._client.request(
            method="GET",
            url=self._base_url / str(data_id),
            auth_required=True,
        ) as resp:
            return await resp.json(loads=absolidix_json_decoder)



"""Low level http and SSE client"""

import json
from asyncio import CancelledError, sleep
from asyncio import TimeoutError as AsyncioTimeoutError
from collections.abc import Awaitable, Callable
from concurrent.futures import TimeoutError as FuturesTimeoutError
from datetime import timedelta
from typing import Any, Optional, Union

import aiohttp
from aiohttp import ClientResponse, ClientTimeout, RequestInfo
from aiohttp.client import (
    _RequestContextManager,  # pyright: ignore [reportPrivateUsage]
)
from aiohttp.client_exceptions import (
    ClientConnectionError,
    ClientPayloadError,
    ClientResponseError,
)
from aiohttp.typedefs import LooseHeaders
from aiohttp.web_exceptions import (
    HTTPInternalServerError,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from aiohttp_sse_client import client as sse_client
from yarl import URL, Query

from .compat import TypedDict, Unpack
from .const import HttpMethods
from .exc import AbsolidixConnectionException, AbsolidixException
from .helpers import (
    JSONT,
    absolidix_json_decoder,
    absolidix_json_encoder,
    http_to_absolidix_error_map,
)
from .models import AbsolidixBase, AbsolidixNoAuth, BaseAuthenticator


class AiohttpSessionRequestOptionsBase(TypedDict, total=False):
    params: Query
    json: Any
    data: Any
    headers: Union[LooseHeaders, None]


class AiohttpSessionRequestOptions(AiohttpSessionRequestOptionsBase, total=False):
    "Partial implementation of the private aiohttp.client._RequestOptions"

    timeout: "Union[ClientTimeout, None]"
    raise_for_status: Union[None, bool, Callable[[ClientResponse], Awaitable[None]]]


class ClientRequestKwargs(AiohttpSessionRequestOptionsBase, total=False):
    "AbsolidixClient.create kwargs"

    method: HttpMethods
    timeout: float
    auth_required: bool


class AbsolidixClient(AbsolidixBase):
    """
    Client to handle API calls.
    Don't use this directly, use `absolidix_client.absolidix.Absolidix` instead to get the client.
    """

    _auth: BaseAuthenticator
    _base_url: URL

    def __init__(
        self,
        session: aiohttp.ClientSession,
        base_url: URL,
        auth: Optional[BaseAuthenticator] = None,
    ) -> None:
        """
        Initialize the Absolidix API client.
        **Arguments:**
        `session`: The aiohttp client session to use for making requests.
        `base_url`: Root URL in form of `yarl.URL`.
        `auth`: Authenticator, subclass of `BaseAuthenticator`.
        """
        self._session = session
        if self._session.json_serialize is not absolidix_json_encoder:
            self._session._json_serialize = absolidix_json_encoder  # pyright: ignore[reportPrivateUsage]
        self._auth = auth or AbsolidixNoAuth()
        if not base_url.is_absolute():
            raise TypeError("Base URL should be absolute")
        self._base_url = base_url

    def _url_rel_to_abs(self, url: URL) -> URL:
        "If url is relative, join it with base url, else return as is"
        if url.is_absolute():
            return url
        if url.path.startswith("/"):
            return self._base_url.join(url)
        return (self._base_url / url.path).with_query(url.query)

    async def _do_auth(self, force: bool = False) -> None:
        async with self._auth.lock:
            if force or await self._auth.should_update(self._session, self._base_url):
                await self._auth.authenticate(self._session, self._base_url)

    @staticmethod
    def _raise_for_status(
        status: int, req_info: RequestInfo, msg: Optional[str] = None
    ) -> None:
        if status < 400:
            return
        msg = f"Response exception for {req_info.url!s} with code {status}: {msg!s}"
        proto_err = http_to_absolidix_error_map(status)
        raise proto_err(status=status, message=msg) from ClientResponseError(
            req_info, message=msg, status=status, history=()
        )

    async def _request(
        self, url: URL, **opts: Unpack[ClientRequestKwargs]
    ) -> ClientResponse:
        """
        Makes an HTTP request to the specified endpoint using the specified parameters.
        """
        url = self._url_rel_to_abs(url)
        method = opts.get("method", "GET")
        auth_required = opts.get("auth_required", False)
        aio_opts: AiohttpSessionRequestOptions = {}
        if "params" in opts:
            aio_opts["params"] = opts["params"]
        if "timeout" in opts:
            aio_opts["timeout"] = ClientTimeout(opts["timeout"])
        if "headers" in opts:
            aio_opts["headers"] = opts["headers"]
        if "json" in opts:
            aio_opts["json"] = opts["json"]
        if "data" in opts:
            aio_opts["data"] = opts["data"]
        aio_opts["raise_for_status"] = False

        # preauthenticate
        if auth_required:
            await self._do_auth()

        try:
            result = await self._session.request(method, url, **aio_opts)

            # redo if failed because of auth
            if result.status == HTTPUnauthorized.status_code:
                # forced auth for first request
                # normal auth (only if needed) for all other
                await self._do_auth(force=not self._auth.lock.locked())
                result.close()
                result = await self._session.request(method, url, **aio_opts)
            # rate limit - redo all
            if result.status == HTTPTooManyRequests.status_code:
                await sleep(10)
                result.close()
                return await self._request(url, **opts)

        except ClientConnectionError as exc:
            raise AbsolidixConnectionException(
                f"Request exception for {str(url)!r} with - {exc}"
            ) from exc

        except (TimeoutError, AsyncioTimeoutError, FuturesTimeoutError):
            raise AbsolidixConnectionException(
                f"Timeout of {opts.get('timeout')} reached while waiting for {str(url)}"
            ) from None

        except BaseException as exc:  # pragma: no cover  # noqa: B036
            raise AbsolidixException(
                f"Unexpected exception for {str(url)!r} with - {exc}"
            ) from exc

        msg = None
        try:
            body = None
            if result.content_type.startswith("application/json"):
                body: JSONT = await result.json(
                    encoding="utf-8",
                    content_type=result.content_type,
                    loads=absolidix_json_decoder,
                )
                if (
                    isinstance(body, dict)
                    and "error" in body
                    and body["error"] is not None
                ):
                    msg = str(body["error"])
            elif result.content_type.startswith("text/"):
                msg = await result.text("utf-8")
            else:
                msg = str(await result.read())
        except (ClientPayloadError, json.JSONDecodeError) as exc:
            raise AbsolidixException(
                f"Broken payload data from {str(url)!r}: {exc}"
            ) from exc
        except BaseException as exc:  # pragma: no cover  # noqa: B036
            raise AbsolidixException(
                f"Could not handle response data from {str(url)!r} with - {exc}"
            ) from exc

        self._raise_for_status(result.status, result.request_info, msg)

        return result

    def request(
        self,
        url: URL,
        **opts: Unpack[ClientRequestKwargs],
    ):
        """
        Makes an HTTP request to the specified endpoint using the specified parameters.
        This method is asynchronous, meaning that it will not block the execution
        of the program while the request is being made and processed.
        **Arguments**:
        -  `url` (Required): The API endpoint to call.
        **Optional arguments**:
        - `data`: The data to include in the request body. Can be a dictionary,
           a string, or None.
        - `headers`: The headers to include in the request. Can be a dictionary or None.
        - `method`: The HTTP method to use for the request. Defaults to GET.
        - `params`: The query parameters to include in the request.
           Can be a dictionary or None.
        - `timeout`: The maximum amount of time to wait for the request to complete,
          in seconds. Can be an integer or None.
        - `auth_required`: Flag that auth requered for this request
        Returns:
        A `aiohttp.client._RequestContextManager` object representing the API response
        """
        return _RequestContextManager(self._request(url, **opts))

    async def sse(
        self,
        url: URL,
        on_message: Callable[[sse_client.MessageEvent], None],
        on_open: Optional[Callable[[], None]] = None,
        params: Optional[Query] = None,
        timeout: Optional[float] = None,
    ) -> None:
        """
        Creates `aiohttp_sse_client.client.EventSource` object with sensible defaults.
        **Arguments**:
        - `url` (Required): The API endpoint to connect.
        - `on_message`: Message callback
        **Optional arguments**:
        - `on_open`: Callback when connected
        - `params`: The query parameters to include in the request.
           Can be a dictionary or None.
        - `timeout`: Stream timeout, None for infinity
        Returns: None
        """
        url = self._url_rel_to_abs(url)
        original_backoff = 0.1
        backoff = original_backoff

        es_timeout = ClientTimeout(total=timeout, sock_connect=600, sock_read=None)
        while True:
            try:
                if backoff > original_backoff:
                    await sleep(backoff)
                await self._do_auth()
                async with sse_client.EventSource(
                    str(url),
                    reconnection_time=timedelta(seconds=original_backoff),
                    max_connect_retry=0,
                    on_open=on_open,
                    session=self._session,
                    params=params,
                    timeout=es_timeout,
                    read_bufsize=2**19,
                    raise_for_status=True,
                ) as evt_src:
                    backoff = original_backoff
                    async for evt in evt_src:
                        on_message(evt)

            except (TimeoutError, AsyncioTimeoutError, FuturesTimeoutError):
                backoff *= 1.5
                self.logger.warning(
                    "%s endpoint timeouted - reconnecting", self._base_url
                )
                continue

            except CancelledError:
                # task is cancelled - abort
                break

            except (ClientConnectionError, ConnectionError) as exc:
                raise AbsolidixConnectionException(
                    f"Connection error for {str(url)!r} with - {exc}"
                ) from exc

            except ClientResponseError as err:
                if err.status and (
                    err.status == HTTPTooManyRequests.status_code
                    or err.status >= HTTPInternalServerError.status_code
                ):
                    backoff *= 2
                    self.logger.warning(
                        "%s connection error %s - reconnecting in %s seconds",
                        url,
                        err,
                        backoff,
                    )
                    continue
                self._raise_for_status(err.status, err.request_info, err.message)



"""Small helpers"""

import json
import re
from collections.abc import Awaitable
from contextlib import suppress
from copy import deepcopy
from datetime import datetime
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union

from aiohttp.web_exceptions import (
    HTTPBadRequest,
    HTTPForbidden,
    HTTPMisdirectedRequest,
    HTTPNotFound,
    HTTPPaymentRequired,
    HTTPUnauthorized,
)

from .compat import ParamSpec, TypeAlias, TypeGuard
from .dtos import (
    AbsolidixErrorDTO,
    AbsolidixErrorEventDataDTO,
    AbsolidixErrorEventDTO,
    AbsolidixErrorMessageDTO,
)
from .exc import (
    AbsolidixAuthenticationException,
    AbsolidixError,
    AbsolidixNotFoundException,
    AbsolidixPayloadException,
    AbsolidixQuotaException,
)

JSONT: TypeAlias = Union[dict[str, "JSONT"], list["JSONT"], str, int, float, bool, None]

LAST_Z_REGEX = re.compile(r"Z$", re.IGNORECASE)

Param = ParamSpec("Param")
RetType = TypeVar("RetType")


def parse_rfc3339(dt_str: Any) -> Optional[datetime]:
    "Parse RFC 3339 date string to datetime object"
    if isinstance(dt_str, datetime):
        return dt_str

    if not dt_str:
        return None

    if dt_str.endswith("Z"):
        dt_str = LAST_Z_REGEX.sub("+00:00", dt_str)

    with suppress(ValueError):
        return datetime.fromisoformat(dt_str)

    for fmt in (
        "%Y-%m-%dT%H:%M:%S.%f%z",
        "%Y-%m-%d %H:%M:%S.%f%z",
        "%Y-%m-%dT%H:%M:%S%z",
        "%Y-%m-%d %H:%M:%S%z",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
    ):
        with suppress(ValueError):
            return datetime.strptime(dt_str, fmt)


_DT = TypeVar("_DT", bound=dict[str, Any])


def convert_dict_values_to_dt(data: _DT) -> _DT:
    "Converts dictionary values to datetime"

    converted = deepcopy(data)
    for key, val in data.items():
        if isinstance(val, str):
            converted[key] = parse_rfc3339(val) or val
        elif isinstance(val, dict):
            converted[key] = convert_dict_values_to_dt(val)  # pyright: ignore[reportUnknownArgumentType]
        elif isinstance(val, list):
            converted[key] = [
                convert_dict_values_to_dt(x) if isinstance(x, dict) else x  # pyright: ignore[reportUnknownArgumentType]
                for x in val  # pyright: ignore[reportUnknownVariableType]
            ]
        elif isinstance(val, tuple):
            converted[key] = tuple(
                convert_dict_values_to_dt(x) if isinstance(x, dict) else x  # pyright: ignore[reportUnknownArgumentType]
                for x in val  # pyright: ignore[reportUnknownVariableType]
            )

    return converted


def convert_dict_values_from_dt(data: _DT) -> _DT:
    "Converts dictionary values from datetime to string"

    converted = deepcopy(data)
    for key, val in data.items():
        if isinstance(val, datetime):
            converted[key] = val.isoformat()
        elif isinstance(val, dict):
            converted[key] = convert_dict_values_from_dt(val)  # pyright: ignore[reportUnknownArgumentType]
        elif isinstance(val, list):
            converted[key] = [
                convert_dict_values_from_dt(x) if isinstance(x, dict) else x  # pyright: ignore[reportUnknownArgumentType]
                for x in val  # pyright: ignore[reportUnknownVariableType]
            ]
        elif isinstance(val, tuple):
            converted[key] = tuple(
                convert_dict_values_from_dt(x) if isinstance(x, dict) else x  # pyright: ignore[reportUnknownArgumentType]
                for x in val  # pyright: ignore[reportUnknownVariableType]
            )

    return converted


def absolidix_json_decoder(
    json_str: str, *args: tuple[Any, ...], **kwargs: dict[str, Any]
):
    "Json decoder but with conversion to snake case and datetime"
    payload: JSONT = json.loads(json_str, *args, **kwargs)
    if isinstance(payload, dict):
        return convert_dict_values_to_dt(payload)
    return payload


def absolidix_json_encoder(
    obj: JSONT, *args: tuple[Any, ...], **kwargs: dict[str, Any]
):
    "Json encoder but with conversion to camel case"
    payload = obj
    if isinstance(obj, dict):
        payload = convert_dict_values_from_dt(obj)
    return json.dumps(payload, *args, **kwargs)


def is_absolidix_error_error_dto(something: Any) -> TypeGuard[AbsolidixErrorMessageDTO]:
    "AbsolidixErrorMessageDTO type guard"
    return (
        isinstance(something, dict)
        and "message" in something
        and isinstance(something["message"], str)
    )


def is_absolidix_error_dto(something: Any) -> TypeGuard[AbsolidixErrorDTO]:
    "AbsolidixErrorDTO type guard"
    return (
        isinstance(something, dict)
        and "status" in something
        and "error" in something
        and (
            isinstance(something["error"], str)
            or is_absolidix_error_error_dto(something["error"])
        )
    )


def is_absolidix_error_event_data_dto(
    something: Any,
) -> TypeGuard[AbsolidixErrorEventDataDTO]:
    "AbsolidixErrorEventDataDTO type guard"
    return (
        isinstance(something, dict)
        and "req_id" in something
        and "data" in something
        and isinstance(something["data"], list)
        and all(is_absolidix_error_dto(x) for x in something["data"])  # pyright: ignore[reportUnknownVariableType]
    )


def is_absolidix_errors_evt_dto(something: Any) -> TypeGuard[AbsolidixErrorEventDTO]:
    "AbsolidixEventDTO type guard"
    return (
        isinstance(something, dict)
        and "type" in something
        and isinstance(something["type"], str)
        and something["type"] == "errors"
        and "data" in something
        and is_absolidix_error_event_data_dto(something["data"])
    )


def http_to_absolidix_error_map(status: int) -> type[AbsolidixError]:
    "Map HTTP exception to AbsolidixError"
    err = AbsolidixError
    if status in (HTTPForbidden.status_code, HTTPUnauthorized.status_code):
        err = AbsolidixAuthenticationException
    if status == HTTPNotFound.status_code:
        err = AbsolidixNotFoundException
    if status == HTTPBadRequest.status_code:
        err = AbsolidixPayloadException
    if status == HTTPPaymentRequired.status_code:
        err = AbsolidixQuotaException
    if status == HTTPMisdirectedRequest.status_code:
        err = AbsolidixError
    return err


def absolidix_error_to_raise(dto: AbsolidixErrorDTO):
    "Raise AbsolidixErrorDTO"
    if isinstance(dto["error"], str):
        msg = dto["error"]
    else:
        msg = dto["error"]["message"]
    err = http_to_absolidix_error_map(dto.get("status"))
    raise err(status=dto["status"], message=msg)


def raise_on_absolidix_error(
    func: Callable[Param, Awaitable[RetType]],
) -> Callable[Param, Awaitable[RetType]]:
    """
    Method wrapper. If there is an AbsolidixErrorDTO in result, than raise AbsolidixError.
    """

    @wraps(func)
    async def wrapped(*args: Any, **kwargs: Any):
        result = await func(*args, **kwargs)
        if is_absolidix_error_dto(result):
            absolidix_error_to_raise(result)

        return result

    return wrapped


def raise_on_absolidix_error_in_event(
    func: Callable[Param, Awaitable[RetType]],
) -> Callable[Param, Awaitable[RetType]]:
    "Raise on AbsolidixErrorDTO in AbsolidixEventDTO"

    @wraps(func)
    async def wrapped(*args: Any, **kwargs: Any):
        result = await func(*args, **kwargs)
        if is_absolidix_errors_evt_dto(result):
            errors = result.get("data", {}).get("data", [])
            if errors:
                absolidix_error_to_raise(errors[-1])

        return result

    return wrapped



"Test AbsolidixError"

from absolidix_client.exc import AbsolidixError


def test_error_string():
    "Test AbsolidixError"
    status = -1234
    message = "mEsSaGe"
    err = AbsolidixError(status=status, message=message)
    assert str(status) in str(err), "Status should be printed"
    assert str(message) in str(err), "Message should be printed"



"Test AbsolidixAPIAsync"

import pytest

from absolidix_client import AbsolidixAPIAsync, AbsolidixNoAuth


async def test_relative_url():
    "Test relative url"
    with pytest.raises(TypeError):
        AbsolidixAPIAsync("/relative", auth=AbsolidixNoAuth())



"""SSE models"""

import json
from dataclasses import dataclass
from functools import partial
from typing import Optional

from aiohttp_sse_client.client import MessageEvent

from ..dtos import (
    AbsolidixCalculationsEventDTO,
    AbsolidixCollectionsEventDTO,
    AbsolidixDataSourcesEventDTO,
    AbsolidixErrorDTO,
    AbsolidixErrorEventDTO,
    AbsolidixEventDTO,
    AbsolidixPongEventDTO,
)
from ..helpers import absolidix_json_decoder


@dataclass(frozen=True)
class AbsolidixMessageEvent(MessageEvent):
    "SSE Message Event"

    type: Optional[str]
    message: str
    data: str
    origin: str
    last_event_id: str

    @classmethod
    def from_dto(cls, data: MessageEvent) -> "AbsolidixMessageEvent":
        "Create model from DTO"
        return cls(
            type=data.type,
            message=data.message,
            data=data.data,
            origin=data.origin,
            last_event_id=data.last_event_id,
        )

    def to_dto(self) -> AbsolidixEventDTO:
        "Create DTO from model"
        try:
            if self.type == "ping":
                return AbsolidixPongEventDTO(type="ping", data="pong")
            for dto in (
                partial(AbsolidixErrorEventDTO, type="errors"),
                partial(AbsolidixDataSourcesEventDTO, type="datasources"),
                partial(AbsolidixCalculationsEventDTO, type="calculations"),
                partial(AbsolidixCollectionsEventDTO, type="collections"),
            ):
                if dto.keywords["type"] in [self.type, self.message]:
                    data = absolidix_json_decoder(self.data)

                    return dto(data=data)
        except json.JSONDecodeError as err:
            return AbsolidixErrorEventDTO(
                type="errors",
                data={
                    "req_id": "",
                    "data": [AbsolidixErrorDTO(status=400, error=str(err))],
                },
            )
        message = f"Unknown event type: {str(self)}"
        return AbsolidixErrorEventDTO(
            type="errors",
            data={"req_id": "", "data": [AbsolidixErrorDTO(status=400, error=message)]},
        )



"""Main Absolidix API async client"""

from types import TracebackType
from typing import Literal, Optional, Union

import aiohttp
from aiohttp import ClientSession, ClientTimeout, TraceConfig
from aiohttp.hdrs import USER_AGENT
from aiohttp.typedefs import LooseHeaders, StrOrURL
from yarl import URL

from .client import AbsolidixClient
from .compat import NotRequired, TypedDict, Unpack
from .const import DEFAULT_USER_AGENT
from .models import AbsolidixBase, BaseAuthenticator
from .namespaces.calculations import AbsolidixCalculationsNamespace
from .namespaces.root import AbsolidixRootNamespace
from .namespaces.stream import AbsolidixStreamNamespace
from .namespaces.v0 import AbsolidixV0Namespace


class AbsolidixAPIKwargs(TypedDict):
    "AbsolidixAPI init kwargs"

    auth: BaseAuthenticator
    headers: NotRequired[LooseHeaders]
    timeout: NotRequired[Union[float, Literal[False], None]]
    client_name: NotRequired[str]
    trace_configs: NotRequired[list[TraceConfig]]


class AbsolidixAPIAsync(AbsolidixBase):
    """Main Absolidix API async client"""

    _close_session = False
    _session: aiohttp.ClientSession

    def __init__(
        self,
        base_url: StrOrURL,
        session: Optional[ClientSession] = None,
        **opts: Unpack[AbsolidixAPIKwargs],
    ):
        """
        Initialize async Absolidix client.

        **Arguments**:

        `bese_url`
        URL of Absolidix BFF server. `str` or `yarl.URL`.

        `session` (Optional)
        `aiohttp.ClientSession` to be used by this package.
        If you do not pass one, one will be created for you.

        `headers` (Optional)
        Optional dictionary of always sent headers. Used if `session` is omitted.

        `timeout` (Optional)
        Optional float timeout in seconds. None if not timeout. False if not set.
        Used as `aiohttp` timeout if `session` is omitted.

        `client_name` (Optional)
        Optional string for user agent.
        Used if `session` is omitted.
        """
        headers = opts.get("headers")
        if session is None:
            timeout = opts.get("timeout", None)
            session = aiohttp.ClientSession(
                timeout=(
                    timeout
                    if isinstance(timeout, ClientTimeout)
                    else ClientTimeout(total=timeout or None)
                ),
                headers=headers,
                trace_configs=opts.get("trace_configs"),
            )
            session.headers[USER_AGENT] = opts.get("client_name", DEFAULT_USER_AGENT)
            self._close_session = True
        self._session = session

        base_url = URL(base_url)
        if not base_url.is_absolute():
            raise TypeError("Base URL should be absolute")
        client = AbsolidixClient(session, base_url, opts["auth"])
        self._ns_root = AbsolidixRootNamespace(client, base_url)

    @property
    def calculations(
        self,
    ) -> AbsolidixCalculationsNamespace:
        """Property to access the calculations namespace."""
        return self._ns_root.calculations

    @property
    def v0(self) -> AbsolidixV0Namespace:
        """Property to access the v0 namespace."""
        return self._ns_root.v0

    @property
    def stream(self) -> AbsolidixStreamNamespace:
        """Property to access the stream namespace."""
        return self._ns_root.stream

    async def __aenter__(self) -> "AbsolidixAPIAsync":
        """Async enter."""
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def close(self) -> None:
        "Close stream and http session"
        self.stream.close()
        if self._session and self._close_session:
            await self._session.close()



"""Data sources DTOs"""

from collections.abc import Sequence
from enum import Enum

from ..compat import TypedDict
from .base import AbsolidixTimestampsDTO
from .collection import AbsolidixCollectionDTO


class DataSourceType(int, Enum):
    """The basic data types supported"""

    STRUCTURE = 1
    CALCULATION = 2
    PROPERTY = 3
    WORKFLOW = 4
    PATTERN = 5
    USER_INPUT = 6


class AbsolidixDataSourceDTO(AbsolidixTimestampsDTO):
    """A basic data item definition"""

    id: int
    parents: Sequence[int]
    children: Sequence[int]
    user_id: int
    user_first_name: str
    user_last_name: str
    user_email: str
    name: str
    content: str
    type: int
    collections: Sequence[AbsolidixCollectionDTO]


class AbsolidixDataSourceContentOnlyDTO(TypedDict):
    """Helper definition"""

    content: str



"Test parse_rfc3339 helper"

from datetime import datetime, timedelta, timezone

import pytest
from typing_extensions import Union

from absolidix_client.helpers import parse_rfc3339

expected_dt = datetime(2001, 2, 3, 4, 5, 6)
expected_dt_tz = expected_dt.replace(tzinfo=timezone.utc)
expected_dt_ms_tz = expected_dt_tz + timedelta(microseconds=789012)

testdata = [
    (None, None),
    ("", None),
    (expected_dt, expected_dt),
    ("2001-02-03 04:05:06Z", expected_dt_tz),
    ("2001-02-03T04:05:06Z", expected_dt_tz),
    ("2001-02-03 04:05:06.789012+00:00", expected_dt_ms_tz),
    ("2001-02-03T04:05:06.789012+00:00", expected_dt_ms_tz),
    ("2001-02-03 04:05:06+00:00", expected_dt_tz),
    ("2001-02-03T04:05:06+00:00", expected_dt_tz),
    ("2001-02-03 04:05:06", expected_dt),
    ("2001-02-03T04:05:06", expected_dt),
]


@pytest.mark.parametrize("x,expected", testdata)
def test_parse_rfc3339(
    x: Union[datetime, str, None], expected: Union[datetime, str, None]
):
    "Test parse_rfc3339"
    assert parse_rfc3339(x) == expected, "Value should be expected"



"""Calculation DTOs"""

from collections.abc import Sequence

from ..compat import NotRequired
from .base import AbsolidixTimestampsDTO
from .datasource import AbsolidixDataSourceDTO


class AbsolidixCalculationDTO(AbsolidixTimestampsDTO):
    "Calculation DTO"

    id: int
    name: str
    user_id: int
    progress: int
    parent: int
    result: NotRequired[Sequence[AbsolidixDataSourceDTO]]



"""Absolidix API client"""

from .absolidix import AbsolidixAPI
from .absolidix_async import AbsolidixAPIAsync
from .const import PROJECT_VERSION as __version__
from .dtos import AbsolidixErrorDTO
from .exc import AbsolidixError
from .models import AbsolidixLocalUserAuth, AbsolidixNoAuth, AbsolidixTokenAuth

__all__ = [
    "AbsolidixAPI",
    "AbsolidixAPIAsync",
    "AbsolidixError",
    "AbsolidixErrorDTO",
    "AbsolidixLocalUserAuth",
    "AbsolidixNoAuth",
    "AbsolidixTokenAuth",
    "__version__",
]
